using BinaryProvider # requires BinaryProvider 0.3.0 or later

# Parse some basic command-line arguments
const verbose = "--verbose" in ARGS
const prefix = Prefix(get([a for a in ARGS if a != "--verbose"], 1, joinpath(@__DIR__, "usr")))
products = [
    ExecutableProduct(prefix, "xc-threshold", :xcthreshold),
    LibraryProduct(prefix, ["libxc"], :libxc),
    ExecutableProduct(prefix, "xc-info", :xcinfo),
]

# Download binaries from hosted location
bin_prefix = "https://github.com/unkcpz/LibxcBuilder/releases/download/v4.3.4"

# Listing of files generated by BinaryBuilder:
download_info = Dict(
    Linux(:i686, libc=:glibc) => ("$bin_prefix/LibxcBuilder.v4.3.4.i686-linux-gnu.tar.gz", "e2b80a859f7a80693a740815293a875714ac3b3261d53f5e58bd57f4021e26d5"),
    MacOS(:x86_64) => ("$bin_prefix/LibxcBuilder.v4.3.4.x86_64-apple-darwin14.tar.gz", "a31ef886b941c72d0ddd96c5ab4950e6b3e03345fd3c78fdab53a7b836bc809c"),
    Linux(:x86_64, libc=:glibc) => ("$bin_prefix/LibxcBuilder.v4.3.4.x86_64-linux-gnu.tar.gz", "8b977256ff2b78280d25b50ac055e86e7cbc0ae339709ddab2a08f3c37f3b5e5"),
)

# Install unsatisfied or updated dependencies:
unsatisfied = any(!satisfied(p; verbose=verbose) for p in products)
dl_info = choose_download(download_info, platform_key_abi())
if dl_info === nothing && unsatisfied
    # If we don't have a compatible .tar.gz to download, complain.
    # Alternatively, you could attempt to install from a separate provider,
    # build from source or something even more ambitious here.
    error("Your platform (\"$(Sys.MACHINE)\", parsed as \"$(triplet(platform_key_abi()))\") is not supported by this package!")
end

# If we have a download, and we are unsatisfied (or the version we're
# trying to install is not itself installed) then load it up!
if unsatisfied || !isinstalled(dl_info...; prefix=prefix)
    # Download and install binaries
    install(dl_info...; prefix=prefix, force=true, verbose=verbose)
end

# Write out a deps.jl file that will contain mappings for our products
write_deps_file(joinpath(@__DIR__, "deps.jl"), products, verbose=verbose)
